/*-
 * Copyright 2015-2016 UPLEX - Nils Goroll Systemoptimierung
 * All rights reserved.
 *
 * Authors: Nils Goroll <nils.goroll@uplex.de>
 *          Geoffrey Simmons <geoffrey.simmons@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

#include "vcl.h"
#include "vrt.h"

#include "vcc_if.h"
#include "vmod_convert.h"
#include "wb.h"

static const struct vmod_priv const null_blob[1] =
{
	{
		.priv = "",
		.len = 0,
		.free = NULL
	}
};

static inline VCL_STRING
concat_va(struct ws *ws, const char *p, va_list ap)
{
	const char *next;
	struct wb_s wb;

	CHECK_OBJ_NOTNULL(ws, WS_MAGIC);
	if (p == vrt_magic_string_end)
		return "";

	SKIP_EMPTY(next, ap);

	if (next == vrt_magic_string_end) {
		if ( p == NULL || *p == '\0')
			return "";
		else
			return p;
	}

	if (!wb_create(ws, &wb) || !wb_append(&wb, p, -1))
		return NULL;

	do {
		if (!wb_append(&wb, next, -1)) {
			wb_reset(&wb);
			return NULL;
		}
		SKIP_EMPTY(next, ap);
	} while (next != vrt_magic_string_end);

	if (wb_len(&wb) == 0) {
		wb_reset(&wb);
		return "";
	}
	return wb_finish(&wb, NULL);
}

static inline size_t
decode_l_va(enum encoding dec, const char * const p, va_list ap)
{
	size_t len = 0;
	const char *next;

	if (p == vrt_magic_string_end)
		return 0;
	if (p != NULL)
		len = strlen(p);

	while ((next = va_arg(ap, const char *)) != vrt_magic_string_end)
		if (next != NULL)
			len += strlen(next);

	switch(dec) {
	case IDENTITY:
		return len;
	case BASE64:
	case BASE64URL:
	case BASE64URLNOPAD:
		return base64_decode_l(len);
	case HEX:
	case HEXLC:
		return hex_decode_l(len);
	default:
		WRONG("Illegal decoding");
	}
}

static inline ssize_t __match_proto__(decode_f)
decode(const enum encoding dec, char *restrict const buf,
       const size_t maxlen, const char *restrict const p, va_list ap)
{
	switch(dec) {
	case IDENTITY:
		return id_decode(dec, buf, maxlen, p, ap);
	case BASE64:
	case BASE64URL:
	case BASE64URLNOPAD:
		return base64_decode(dec, buf, maxlen, p, ap);
	case HEX:
		return hex_decode(dec, buf, maxlen, p, ap);
	default:
		WRONG("Illegal decoding");
	}
}

static inline ssize_t __match_proto__(encode_f)
encode(const enum encoding enc, char *restrict const buf, const size_t maxlen,
       const char *restrict const in, const size_t inlen)
{
	switch(enc) {
	case IDENTITY:
		return id_encode(enc, buf, maxlen, in, inlen);
	case BASE64:
	case BASE64URL:
	case BASE64URLNOPAD:
		return base64_encode(enc, buf, maxlen, in, inlen);
	case HEX:
	case HEXLC:
		return hex_encode(enc, buf, maxlen, in, inlen);
	default:
		WRONG("Illegal encoding");
	}
}

/* init / event handler */
int __match_proto__(vmod_event_f)
event(VRT_CTX, struct vmod_priv *priv, enum vcl_event_e event)
{
	(void) priv;

	if (event != VCL_EVENT_LOAD)
		return 0;

	base64_init();
	return (0);
}

/* Functions */

VCL_BLOB __match_proto__(td_convert_decode)
vmod_decode(VRT_CTX, VCL_ENUM decs, const char *p, ...) {
	enum encoding dec = parse_encoding(decs);
	va_list ap;
	struct wb_s wb;
	struct vmod_priv *b;
	char *snap, *buf;
	ssize_t len;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);

	snap = WS_Snapshot(ctx->ws);
	if ((b = WS_Alloc(ctx->ws, sizeof(struct vmod_priv))) == NULL)
		return NULL;
	if (wb_create(ctx->ws, &wb) == NULL) {
		WS_Reset(ctx->ws, snap);
		return NULL;
	}
	buf = wb_buf(&wb);

	va_start(ap, p);
	len = decode(dec, buf, wb_space(&wb), p, ap);
	va_end(ap);

	if (len == -1) {
		wb_reset(&wb);
		WS_Reset(ctx->ws, snap);
		return NULL;
	}
	if (len == 0) {
		wb_reset(&wb);
		WS_Reset(ctx->ws, snap);
		return null_blob;
	}
	wb_advance(&wb, len);
	WS_ReleaseP(ctx->ws, wb_buf(&wb));
	b->priv = buf;
	b->len = len;
	b->free = NULL;
	return b;
}

VCL_STRING __match_proto__(td_convert_encode)
vmod_encode(VRT_CTX, VCL_ENUM encs, VCL_BLOB b) {
	enum encoding enc = parse_encoding(encs);
	struct wb_s wb;
	ssize_t len;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);
	if (b == NULL)
		return NULL;

	if (wb_create(ctx->ws, &wb) == NULL)
		return NULL;

	len = encode(enc, wb_buf(&wb), wb_space(&wb), b->priv, b->len);

	if (len == -1) {
		wb_reset(&wb);
		return NULL;
	}
	if (len == 0) {
		wb_reset(&wb);
		return "";
	}
	wb_advance(&wb, len);
	return wb_finish(&wb, NULL);
}

VCL_STRING __match_proto__(td_convert_transcode)
vmod_transcode(VRT_CTX, VCL_ENUM decs, VCL_ENUM encs, const char *p, ...) {
	enum encoding dec = parse_encoding(decs);
	enum encoding enc = parse_encoding(encs);
	struct wb_s wb;
	ssize_t len;
	va_list ap;

	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);

	/*
	 * If the encoding and decoding are the same, just return the
	 * concatenated string.
	 * XXX: Doesn't reject illegal decodings.
	 */
	if (enc == dec) {
		va_start(ap, p);
		VCL_STRING s = concat_va(ctx->ws, p, ap);
		va_end(ap);
		return s;
	}

	/* Allocate space for the decoded blob on the stack */
	va_start(ap, p);
	size_t l = decode_l_va(dec, p, ap);
	va_end(ap);
	if (l == 0)
		return "";
	assert(l > 0);
	char buf[l];

	va_start(ap, p);
	len = decode(dec, buf, l, p, ap);
	va_end(ap);

	if (len == -1)
		return NULL;

	if (wb_create(ctx->ws, &wb) == NULL)
		return NULL;

	len = encode(enc, wb_buf(&wb), wb_space(&wb), buf, len);
	if (len == -1) {
		wb_reset(&wb);
		return NULL;
	}
	if (len == 0) {
		wb_reset(&wb);
		return "";
	}
	wb_advance(&wb, len);
	return wb_finish(&wb, NULL);
}
